/*
** Module   : EventDefinition.java
** Abstract : represents events and widgets that determine the condition of
**            termination of WAIT-FOR statements and/or triggers
**
** Copyright (c) 2007-20103, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20070112   @31882 Created initial version.
** 002 NVS 20070417   @33088 Clarification of parameters of the lookup()
**                           method. New static method isEventUniversal().
** 003 SIY 20070517   @33753 Javadoc cleanup.
** 004 GES 20071227   @36594 Added safety code to handle intersection and
**                           difference with an event def that has a null
**                           widget set.
** 005 NVS 20080320   @37618 Added anyEventForWidget() lookup method.
** 006 SIY 20080709   @39084 Inferred generics.
** 007 SIY 20080901   @39643 Added ANY-KEY event support.
** 008 CA  20081204   @40815 Added getRowDisplayEvents, which collects 
**                           all ROW-DISPLAY events.
** 009 CA  20090408   @41732 Added getWidgetIds - returns the widget ids
**                           registered with this event definition.
** 010 GES 20090629   @42968 Added isSame() helper to test equality. Improved
**                           serialization efficiency. 
** 011 SIY 20101122          Inferred generics and other minor cleanups.
** 012 CA  20130822          Added support of appserver async requests. The changes are related to
**                           allow WAIT-FOR <server-event> of <handle> statements to be handled
**                           by P2J. In this case, the resource IDs are kept separately from the
**                           widget IDs (even in case when the handle targets a widget).
** 013 HC  20131031          Changed resource ID data type to Long, see issue #2183.
** 014 EVL 20140808          Rework for the new widget ID implementation. 
*/ 

package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.util.Utils;

/**
 * This class represents events and widgets that determine the condition of
 * termination of implicit or explicit WAIT-FOR statement and triggers 
 * activation.
 * <p>
 * Multiple events can be associated with multiple widgets. Any combination
 * of a single event with a single widget is a valid exit condition or a
 * trigger.
 * <p>
 * If there is no widgets specified, the case corresponds to the ANYWHERE
 * option specified globally. Besides, a specific ANYWHERE option that
 * accompanies the set of widgets can be specified.
 * <p>
 * If the association defines a trigger, the trigger ID should be a valid
 * trigger ID from the trigger registry on the server. A special value of -1
 * means there is no trigger registered with this even definition instance,
 * and, therefore, it defines the WAIT-FOR exit conditions.
 * <p>
 * A single instance of this class represents one association of a group of
 * events with a group of widgets and/or the ANYWHERE option.
 * To define a full list of triggers or exit conditions for a I/O statement,
 * more than one instance may be necessary.
 * <p>
 * Widgets are specified with their numeric IDs. Events are specified with
 * their symbolic names. These measures make this class independent from the
 * server's classes and easily serializable.
 * <p>
 * <p>
 * The targets' IDs will be split in two sets, depending on the resource type:
 * <ol>
 *    <li>The IDs for all widget resources will be added to the {@link EventDefinition#widgets}
 *        set.</li>
 *    <li>The IDs for all non-widget resources will be added to the 
 *         {@link EventDefinition#resources} set.</li> 
 * </ol>
 * The main reason for this approach is that widgets on P2J client-side have their own unique IDs
 * which are computed different from their associated resource ID on server-side. Thus, mixing 
 * client-side widget IDs and resource IDs would not allow to differentiate which is which and it
 * cascades in lot of changes in the UI client code.  Also, for the same reason, a widget 
 * referenced via a handle in a i.e. "wait-for close of h" statement will register it in the set 
 * of widget IDs, and not resources. The other reason was that P2J uses a stack of UI event-queues
 * which is empty when no wait-for loop is running; thus, server-events can not be posted there 
 * because they will get discarded if the event-queue is "trashed".
 */
public class EventDefinition
implements Externalizable
{
   /** Set of event names. Can't be null. */
   private Set<String> events = new HashSet<String>();

   /** Set of widget IDs or null if anywhere globally. */
   private Set<Integer> widgets = null;

   /** Set of non-widget resource IDs or null if anywhere globally. */
   private Set<Long> resources = null;

   /** Specific ANYWHERE option, if given. */
   private boolean anywhere = false;

   /** Trigger assigned to the event definition or -1 if exit condition. */
   private int triggerId = -1;

   /**
    * Default constructor.
    */
   public EventDefinition()
   {
   }
   
   /**
    * A constructor.
    *
    * @param   events
    *          a Set of event names
    * @param   widgets
    *          a Set of widget IDs
    * @param   resources
    *          a Set of non-widget resource IDs
    * @param   anywhere
    *          <code>true</code> if these events are defined as ANYWHERE
    */
   public EventDefinition(Set<String>  events,
                          Set<Integer> widgets,
                          Set<Long> resources,
                          boolean      anywhere)
   {
      this.events    = events;
      this.widgets   = widgets;
      this.resources = resources;
      this.anywhere  = anywhere;
   }
   
   /**
    * Classifies events as either universal or regular.
    *
    * @param   eventCode
    *          numeric event code
    *          
    * @return  <code>true</code> if event is universal.
    */
   public static boolean isEventUniversal(int eventCode)
   {                              
      // certain actions are honored as if the event definition was anywhere
      // in the frame, even when no anywhere flag was specified by the
      // programmer (these are implicitly anywhere across the entire frame)
      boolean universal = (eventCode == Keyboard.KA_GO        ||
                           eventCode == Keyboard.KA_ERROR     ||
                           eventCode == Keyboard.KA_ENDKEY    ||
                           eventCode == Keyboard.KA_END_ERROR ||
                           eventCode == Keyboard.KA_STOP);
      
      return universal;
   }
   
   /**              
    * Reports if the given event definition is the same (in all data except for the trigger ID) 
    * as the current instance.
    *
    * @param    other
    *           The event definition to compare against.
    *
    * @return   <code>true</code> if the definitions are the same except the trigger ID.
    */
   public boolean isSame(EventDefinition other)
   {
      if (other != null)
      {
         if (anywhere != other.anywhere)
            return false;
         
         if (events.size() != other.events.size() ||
             !events.containsAll(other.events))
            return false;
            
         return Utils.isSame(widgets, other.widgets) && Utils.isSame(resources, other.resources);
      }
         
      return false;
   }

   /**
    * Get the widget ids registered for this event.
    * 
    * @return   the widget id list registered for this event. may be null, in
    *           case of anywhere globally.
    */
   public Set<Integer> getWidgetIds()
   {
      return widgets;
   }
   
   /**
    * Get the resource ids registered for this event.
    * 
    * @return   the resource id list registered for this event. May be null, in case of anywhere 
    *           globally.
    */
   public Set<Long> getResourceIds()
   {
      return resources;
   }

   /**
    * Search for registered ROW-DISPLAY events.
    * 
    * @return   a map containing pairs of (widgetId, triggerId). may be null
    */
   public Map<Integer, Integer> getRowDisplayEvents()
   {
      Map<Integer, Integer> res = null;
      
      if (events.contains(Keyboard.ROW_DISPLAY) && 
          (anywhere || 
           widgets != null))
      {
         res = new HashMap<Integer, Integer>();
         
         if (anywhere)
         {
            res.put(-1, triggerId);
         }
         
         if (widgets != null)
         {
            for (int wid : widgets)
            {
               res.put(wid, triggerId);
            }
         }
      }
      
      return res;
   }

   /**
    * Sets trigger ID for this event definition making it a trigger event
    * definition.
    *
    * @param   triggerId
    *          ID of a trigger from the trigger registry
    */
   public void setTriggerId(int triggerId)
   {
      this.triggerId = triggerId;
   }
   
   /**
    * Gets trigger ID of this event definition.
    *
    * @return  an ID of a trigger or <code>-1</code> if this event definition
    *          is for an exit condition.
    */
   public int getTriggerId()
   {
      return triggerId;
   }
   
   /**
    * Checks whether this event matches any event defined in this
    * <code>EventDefinition</code> instance.
    *
    * @param    frameId
    *           Frame ID, can be <code>-1</code>
    * @param    widgetId
    *           widget ID to check or -1 if global anywhere event
    * @param    event
    *           event name in a proper case (the caller should take care)
    * @param    trigExit
    *           flag, telling the search is for a trigger (<code>true</code>)
    *           or for an exit condition (<code>false</code>).
    * @param    specific
    *           <code>true</code> if an exact match is to be returned
    * @param    universal
    *           <code>true</code> if the event is of the universal class
    * @param    isKey
    *           This flags marks reals keys typed by user.
    * @return  returns trigger ID or -1 for triggers;
    *          returns 0 or -1 for exit conditions. In either case,
    *          -1 means no match was found.
    */
   public int lookup(int     frameId,
                     int     widgetId,
                     String  event,
                     boolean trigExit, 
                     boolean specific,
                     boolean universal, 
                     boolean isKey)
   {
      // check the search type
      if (trigExit && triggerId == -1)
         return -1;                      // no match; not a trigger

      if (!trigExit && triggerId != -1)
         return -1;                      // no match; not an exit condition
      
      // check if the event is listed
      if (!events.contains(event))
      {
         if (!isKey || !events.contains("ANY-KEY"))
            return -1;
      }
      
      // check global anywhere case
      if (widgets == null && widgetId == -1)
      {
         return trigExit ? triggerId : 0;
      }

      if (widgets == null && widgetId != -1)
         return -1;                      // no match; only global anywhere

      if (widgetId == -1)
         return -1;                      // no match; only specific widgets

      // check specific widget
      boolean match = widgets.contains(new Integer(widgetId));
      if (match)
      {
         return trigExit ? triggerId : 0;
      }
      
      if (specific)
         return -1;
          
      // check anywhere within the frame
      if (!anywhere && !universal)
         return -1;
      
      // this checking can be done on client and server side with single call
      if (WidgetId.possibleValidFrameId(widgetId))
         return -1;
      
      int fId = -1;
      
      // determine if the calling environment is the client side or server side
      if (ThinClient.getInstance() != null)
      {  // got client session we are on the client side
         fId = ThinClient.frameIdFromWidgetId(widgetId);
      }
      else
      {   // not a client considered as server side
         fId = LogicalTerminal.frameIdFromWidgetId(widgetId);
      }
      
      if (fId != frameId)
         return -1;
      
      match = widgets.contains(new Integer(frameId));
      if (match)
      {
         return trigExit ? triggerId : 0;
      }
      
      return -1;
   }

   /**
    * Checks whether this server event matches any server event defined in this
    * <code>EventDefinition</code> instance. Only exit conditions are checked.
    *
    * @param    resourceId
    *           The ID of the resource.
    * @param    event
    *           event name in a proper case (the caller should take care)
    *
    * @return   returns 0 if a match was found or -1 when no match was found.
    */
   public int lookup(long resourceId, String event)
   {
      // check if the event is listed
      if (!events.contains(event))
      {
         return -1;
      }
      
      // check if there are non-widget resources registered in this event definition
      if (resources == null)
      {
         return -1;
      }

      return resources.contains(resourceId) ? 0 : -1;
   }
   
   /**
    * Checks whether this definition contains exit conditions referring
    * to the specified widget.
    *
    * @param    widgetId
    *           widget ID to check
    * 
    * @return   <code>true</code> if the widget is referred from here
    */
   public boolean anyEventForWidget(int widgetId)
   {
      if (triggerId != -1)
      {
         return false;                      // not an exit condition
      }

      // check global anywhere case
      if (widgets == null)
      {
         return false;                      // not an exit condition
      }

      // check specific widget
      return widgets.contains(new Integer(widgetId));
   }

   /**
    * Checks whether this event definition has common events, widgets and resources with the
    * specified one.
    * <p>
    * They intersect, if:
    * <ul>
    *   <li>event sets intersect;
    *   <li>either both widget and resource sets intersect or don't exist.
    * </ul>
    *
    * @param    ed
    *           another instance of <code>EventDefinition</code>
    * 
    * @return   <code>true</code> if the event definitions have common events and widgets
    */
   public boolean intersects(EventDefinition ed)
   {
      if (!Utils.intersect(events, ed.events))
         return false;
      
      if ((widgets == null && ed.widgets != null) ||
          (widgets != null && ed.widgets == null) ||
          (resources != null && ed.resources == null) ||
          (resources == null && ed.resources != null))
         return false;
         
      return ((widgets == null && ed.widgets == null) || 
               Utils.intersect(widgets, ed.widgets)) &&
             ((resources == null && ed.resources == null) || 
               Utils.intersect(resources, ed.resources));
   }
   
   /**
    * Returns an instance of <code>EventDefinition</code> which represents the common events, 
    * widgets and resources with the specified one.
    * <p>
    * The returned <code>EventDefinition</code> retains all remaining internal fields of this one.
    *
    * @param    ed
    *           another instance of <code>EventDefinition</code>
    * 
    * @return   common part as a <code>EventDefinition</code> or <code>null</code> if there is 
    *           nothing in common.
    */
   public EventDefinition intersection(EventDefinition ed)
   {
      if (!intersects(ed))
         return null;
      
      EventDefinition ned = new EventDefinition();
      ned.anywhere  = anywhere;
      ned.triggerId = triggerId;
      ned.events    = new HashSet<String>();
      
      Utils.intersection(events, ed.events, ned.events);
      
      // if either or both of the widget sets is not null, then the result
      // is not null (create an empty set)
      if (widgets != null || ed.widgets != null)
      {
         ned.widgets = new HashSet<Integer>();
      }
      
      // if both are not null, calculate the intersection (the ned.widgets
      // is an empty set at this point)
      if (widgets != null && ed.widgets != null)
      {
         Utils.intersection(widgets, ed.widgets, ned.widgets);
      }
      
      // if either or both of the resource sets is not null, then the result
      // is not null (create an empty set)
      if (resources != null || ed.resources != null)
      {
         ned.resources = new HashSet<Long>();
      }
      
      // if both are not null, calculate the intersection (the ned.widgets
      // is an empty set at this point)
      if (resources != null && ed.resources != null)
      {
         Utils.intersection(resources, ed.resources, ned.resources);
      }

      return ned;
   }
   
   /**
    * Returns a list of 0 to 2 instances of <code>EventDefinition</code>s which represent the 
    * events, widgets and resources of this instance which are not in the specified one.
    * <p>
    * The result list is empty if event, widget and resource lists match exactly.
    * <p>
    * If this is not the case, the result list contains 1 instance when widget lists match in both
    * instances or don't exist, otherwise two.
    * <p>
    * More specifically, the resulting list is calculated as follows:
    * <ul>
    *   <li>intersection of event sets IE is calculated;
    *   <li>difference of event sets DE is calculated;
    *   <li>intersection of widget sets IW is calculated;
    *   <li>difference of widget sets DW is calculated;
    *   <li>intersection of resource sets IR is calculated;
    *   <li>difference of resource sets DR is calculated;
    *   <li>the first instance of <code>EventDefinition</code> is created off the DE, IW and IR 
    *       sets;
    *   <li>the second instance of <code>EventDefinition</code> is created off this events 
    *       (original set), and DW and DR sets if either DW or DR set is not empty
    * </ul>
    *
    * @param    ed
    *           another instance of <code>EventDefinition</code>
    * 
    * @return   a list of <code>EventDefinition</code>s or <code>null</code>
    *           if there is nothing in common.
    */
   @SuppressWarnings("null")
   public LinkedList<EventDefinition> difference(EventDefinition ed)
   {
      if (!intersects(ed))
          return null;
      
      // calculate all sets
      LinkedList<EventDefinition> l = new LinkedList<EventDefinition>();
      HashSet<String> ie = new HashSet<String>();
      HashSet<String> de = new HashSet<String>();
      HashSet<Integer> iw = new HashSet<Integer>();
      HashSet<Integer> dw = new HashSet<Integer>();
      HashSet<Long> ir = new HashSet<Long>();
      HashSet<Long> dr = new HashSet<Long>();

      Utils.intersection(events, ed.events, ie);
      Utils.difference(events, ed.events, de);

      // if both widget sets are null then the intersection is null and
      // the difference is null
      if (widgets == null && ed.widgets == null)
      {
         iw = null;
         dw = null;
      }
      
      // if our widget set is not null but the other widget set is null then
      // the intersection is an empty set and the difference is the widgets
      // set
      if (widgets != null && ed.widgets == null)
      {
         dw.addAll(widgets);
      }
      
      // if our widget set is null but the other widget set is not null then
      // the intersection and difference are both empty sets
      if (widgets == null && ed.widgets != null)
      {
         // nothing to do
      }
      
      // if both widget sets are not null, calculate the intersection and
      // difference
      if (widgets != null && ed.widgets != null)
      {
         Utils.intersection(widgets, ed.widgets, iw);
         Utils.difference(widgets, ed.widgets, dw);
      }

      // if both resources sets are null then the intersection is null and the difference is null
      if (resources == null && ed.resources == null)
      {
         ir = null;
         dr = null;
      }
      
      // if our resource set is not null but the other resource set is null then the intersection
      // is an empty set and the difference is the resources set
      if (resources != null && ed.resources == null)
      {
         dr.addAll(resources);
      }
      
      // if our resource set is null but the other widget set is not null then the intersection 
      // and difference are both empty sets
      if (resources == null && ed.resources != null)
      {
         // nothing to do
      }
      
      // if both resource sets are not null, calculate the intersection and difference
      if (resources != null && ed.resources != null)
      {
         Utils.intersection(resources, ed.resources, ir);
         Utils.difference(resources, ed.resources, dr);
      }
      
      // conditionally create the first event definition of the result
      if (de.size() > 0)
      {
         EventDefinition ed1 = new EventDefinition();
         ed1.anywhere  = anywhere;
         ed1.triggerId = triggerId;
         ed1.events    = de;
         ed1.widgets   = iw;
         ed1.resources = ir;
         l.add(ed1);
      }

      dw = (dw == null || dw.size() == 0 ? null : dw);
      dr = (dr == null || dr.size() == 0 ? null : dr);
      // conditionally create the second event definition of the result
      if (dw != null || dr != null)
      {
         EventDefinition ed2 = new EventDefinition();
         ed2.anywhere  = anywhere;
         ed2.triggerId = triggerId;
         ed2.events    = events;
         ed2.widgets   = dw;
         ed2.resources = dr;
         l.add(ed2);
      }

      return l;
   }
      
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      // the following members are always present
      Collections.addAll(events, (String[]) in.readObject());      
      anywhere  = in.readBoolean();
      triggerId = in.readInt();
      
      // the rest of the data are objects which can be null so we
      // read an indicator var first to determine if there is following
      // data to be read

      if (in.readBoolean())
      {
         int[] wids = (int[]) in.readObject();
         widgets = new HashSet<Integer>(Utils.primitiveToIntegerList(wids));
      }
      if (in.readBoolean())
      {
         long[] rids = (long[]) in.readObject();
         resources = new HashSet<Long>(Utils.primitiveToLongList(rids));
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      // the following members are always present
      out.writeObject(events.toArray(new String[0]));      
      out.writeBoolean(anywhere);
      out.writeInt(triggerId);
      
      // the rest of the data are objects which can be null so we
      // write an indicator var first to specify if there is following
      // data to be read
      
      boolean there = (widgets != null);
      out.writeBoolean(there);
      
      if (there)
      {
         int[] wids = Utils.integerCollectionToPrimitive(widgets);
         out.writeObject(wids);
      }
      
      there = (resources != null);
      out.writeBoolean(there);
      if (there)
      {
         long[] rids = Utils.longCollectionToPrimitive(resources);
         out.writeObject(rids);
      }
   }
}
