/*
** Module   : WidgetId.java
** Abstract : represents widget identifiers and their relationships
**
** Copyright (c) 2007-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- --------------------------Description----------------------------
** 001 NVS 20070116 NEW   @31884 Created initial version.
** 002 EVL 20140731              Rework to centralize all widget ID related code in single place.
*/ 

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.security.*;

/**
 * This class represents widget identifiers.  Every JVM session has it's own set of widget ID.
 * <p>
 * Approach currently used reserves the 1 number for default window.  Every new widget static or
 * dynamic will consume numbers from 2 up to maximum value of the <code>int</code> variable type.
 * The widget ID is not normally reused when the widget is deleted.  Instead this freed number
 * can be used again when integer number wrapping happen - integer become negative in attempt to
 * increase after maximum value reached.  Two sequential pass of full integer range without
 * getting free ID means there is no free ID and this generates runtime exception to avoid
 * infinite loop.
 */
public class WidgetId
{
   /** The ID for default window widget. */
   private static final int DEFAULT_WINDOW_ID = 1;
   
   /** Starting number for the first widget ID. */
   private static final int WIDGET_ID_RANGE_BEGIN = DEFAULT_WINDOW_ID + 1;
   
   /** Stores session context-local state variables. */
   private static ContextContainer session = new ContextContainer(); 
   
   /**
    * Retrieve the unique per session integer id number for next new widget to create.
    *
    * @return The integer Id number for next dynamic widget to create. 
    *
    */
   public static int allocateID()
   {
      return session.obtain().allocateId();
   }
   
   /**
    * Retrieve the unique per session integer id number for new window widget to create.  Only
    * one window per session possible. 
    *
    * @return The integer Id number for window widget to create. 
    *
    */
   public static int allocateWindowID()
   {
      return DEFAULT_WINDOW_ID;
   }
   
   /**
    * Determines if the given ID could be a valid frame ID (but does not
    * determine if the given ID actually represents a valid frame).
    *
    * @param    fid
    *           Possible frame ID.
    *
    * @return   <code>true</code> if the given ID matches the rules used by 
    *           this class to create frame IDs.
    */
   public static boolean possibleValidFrameId(int fid)
   {
      return fid > DEFAULT_WINDOW_ID;
   }

   /** 
    * Stores global data relating to the state of the current session context.
    */
   private static class SessionArea
   {  
      /** The initial value for next widget ID. */
      private int nextId = WIDGET_ID_RANGE_BEGIN;
      
      /** Flag indicating the integer wrapping happen while incrementing by 1. */
      private boolean wrapped = false;
   
      /**
       * Retrieve the unique per session integer id number for next new widget to create.
       *
       * @return The integer Id number for next dynamic widget to create. 
       *
       * @throws   RuntimeException
       *           If all possible IDs are in use. This is very unlikely but could happen!
       */
      private int allocateId()
      {
         // return result
         int candidate = -1;
         
         while (true)
         {
            // we are deliberately assigning the value before incrementing
            candidate = nextId++;

            if (wrapped)
            {
               // lookup candidate in the widget registry to see if a widget of that ID already
               // exists, don't keep a new widget registry, use the one that already exists

               // if the id is already in use
               if (LogicalTerminal.isWidgetIdInUse(candidate))
               {
                  continue;
               }

               // don't return directly from here
            }

            // protection from infinite loop when there is no free ID to allocate
            // we are here on the second wrapping
            if (nextId < 0 && wrapped)
            {
               throw new RuntimeException("No free ID to allocate!");
            }
            
            if (nextId < 0)
            {
               wrapped = true;
               nextId  = WIDGET_ID_RANGE_BEGIN;
            }

            // at this point we know that candidate is OK so we exit our loop
            break;
         }

         return candidate;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<SessionArea>
   {
      /**
       * Obtains the context-local instance of the contained global work area for the session.
       *
       * @return   The work area associated with this context.
       */
      public SessionArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a new session context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized SessionArea initialValue()
      {
         SessionArea sa = new SessionArea();

         return sa;
      }   
   }   
}

